@@ -81,7 +81,9 @@ typedef struct _Context {
 	unsigned int uiMappedLength;
 	clock_t timeT;
 	unsigned int frameCount;
-	
+
+	GLint maxVertexAttribs;
+	GLboolean* vertexAttribStates;
 } Context;
 
 static const char vshader[] = ""
@@ -164,6 +166,9 @@ static void newContext(Context * ctx) {
 	glAttachShader(ctx->uiProgram, vs);
 	glAttachShader(ctx->uiProgram, fs);
 	glLinkProgram(ctx->uiProgram);
+
+	glGetIntegerv(GL_MAX_VERTEX_ATTRIBS, &ctx->maxVertexAttribs);
+	ctx->vertexAttribStates = (GLboolean*)calloc(ctx->maxVertexAttribs, sizeof(GLboolean));
 }
 
 static void releaseMem(Context *ctx) {
@@ -418,17 +423,9 @@ static void drawOverlay(Context *ctx, unsigned int width, unsigned int height) {
 			    right, top,
 			    right, bottom};
 	GLfloat tex[] = {xm, ymx, xm, ym, xmx, ym, xmx, ymx};
-
-	glBegin(GL_QUADS);
-	glTexCoord2f(tex[0], tex[1]);
-	glVertex2f(vertex[0], vertex[1]);
-	glTexCoord2f(tex[2], tex[3]);
-	glVertex2f(vertex[2], vertex[3]);
-	glTexCoord2f(tex[4], tex[5]);
-	glVertex2f(vertex[4], vertex[5]);
-	glTexCoord2f(tex[6], tex[7]);
-	glVertex2f(vertex[6], vertex[7]);
-	glEnd();
+	glVertexPointer(2, GL_FLOAT, 0, vertex);
+	glTexCoordPointer(2, GL_FLOAT, 0, tex);
+	glDrawArrays(GL_QUADS, 0, 4);
 
 	glPopMatrix();
 }
@@ -542,6 +539,16 @@ static void drawContext(Context * ctx, int width, int height) {
 	glDisable(GL_VERTEX_PROGRAM_ARB);
 	glDisable(GL_FRAGMENT_PROGRAM_ARB);
 
+	GLint enabled;
+	for (i=0;i<ctx->maxVertexAttribs;++i) {
+		enabled = GL_FALSE;
+		glGetVertexAttribiv((GLuint)i, GL_VERTEX_ATTRIB_ARRAY_ENABLED, &enabled);
+		if (enabled == GL_TRUE) {
+			glDisableVertexAttribArray((GLuint)i);
+			ctx->vertexAttribStates[i] = GL_TRUE;
+		}
+	}
+
 	glUseProgram(ctx->uiProgram);
 
 	glEnable(GL_COLOR_MATERIAL);
@@ -559,16 +566,28 @@ static void drawContext(Context * ctx, int width, int height) {
 	glEnableClientState(GL_VERTEX_ARRAY);
 	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
 
-	int bound = 0;
+	int bound = 0, vbobound = 0;
 	glGetIntegerv(GL_PIXEL_UNPACK_BUFFER_BINDING, &bound);
+	glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &vbobound);
 
 	if (bound != 0)
 		glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
+	if (vbobound != 0)
+		glBindBuffer(GL_ARRAY_BUFFER, 0);
 
 	drawOverlay(ctx, width, height);
 
 	if (bound != 0)
 		glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, bound);
+	if (vbobound != 0)
+		glBindBuffer(GL_ARRAY_BUFFER, vbobound);
+
+	for (i=0;i<ctx->maxVertexAttribs;++i) {
+		if (ctx->vertexAttribStates[i] == GL_TRUE) {
+			glEnableVertexAttribArray((GLuint)i);
+			ctx->vertexAttribStates[i] = GL_FALSE;
+		}
+	}
 
 	glMatrixMode(GL_TEXTURE);
 	glPopMatrix();
