@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2010, Thorvald Natvig <thorvald@natvig.com>
+/* Copyright (C) 2005-2011, Thorvald Natvig <thorvald@natvig.com>
 
    All rights reserved.
 
@@ -28,10 +28,11 @@
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
-#ifndef _OSINFO_H
-#define _OSINFO_H
+#ifndef OSINFO_H_
+#define OSINFO_H_
 
-#include "murmur_pch.h"
+class QDomDocument;
+class QDomElement;
 
 class OSInfo {
 	public:
